"""Provides a class that acts as a thin wrapper around Plottext.

Because (as of the time of writing) there doesn't seem to be a method of
getting a plotting object from Plottext, this class pulls a figure out from
inside Plotext and promotes it to the top level, so to speak, and then takes
care of adding in some of the utility methods that will bee needed.
"""

from __future__ import annotations

from typing import Any, Tuple, Union

from typing_extensions import Literal, TypeAlias, get_args

from plotext._dict import (
    themes as _themes,
    type1_to_type2_codes,
)
from plotext._utility import get_color_code

from textual.app import DEFAULT_COLORS
from textual.color import Color as TextualColor

from . import plotext
from .plotext._figure import _figure_class as Figure

ThemeName = Literal[
    # The standard Plotext themes.
    "clear",
    "dark",
    "default",
    "dreamland",
    "elegant",
    "girly",
    "grandpa",
    "matrix",
    "mature",
    "pro",
    "retro",
    "sahara",
    "salad",
    "scream",
    "serious",
    "windows",
    # Full colour versions of the Plotext themes.
    "textual-clear",
    "textual-dark",
    "textual-default",
    "textual-dreamland",
    "textual-elegant",
    "textual-girly",
    "textual-grandpa",
    "textual-matrix",
    "textual-mature",
    "textual-pro",
    "textual-retro",
    "textual-sahara",
    "textual-salad",
    "textual-scream",
    "textual-serious",
    "textual-windows",
    # Textual extra themes.
    "textual-design-dark",
    "textual-design-light",
]
"""Literal type that is the list of theme names defined in Plotext.

Note:
    This is a copy of the theme names and will need to be updated if Plotext
    ever adds more. The main reasons for taking this approach are:

    1. There is no public interface to access this information.
    2. Turning runtime information into type-checking data isn't simple.
"""

Color: TypeAlias = Union[str, int, Tuple[int, int, int]]
"""Type of a Plotext colour."""


class Plot(Figure):
    """A class that provides a Textual-friendly interface to Plotext.

    This class inherits from Plotext's `_figure_class` and then adds access
    to some global functions, making it appear like the top-level `plotext`
    module.

    The aim of the class is to make this:

    ```python
    import plotext as plt

    plt.title("Scatter Plot")
    plt.scatter(plt.sin())
    ```

    and this:

    ```python
    plt = Plot()
    plt.title("Scatter Plot")
    plt.scatter(plt.sin())
    ```

    functionally equivalent, but with the advantage that the latter has no
    global state and is free of external side-effects.
    """

    @staticmethod
    def sin(
        periods: int = 2,
        length: int = 200,
        amplitude: int = 1,
        phase: float = 0,
        decay: int = 0,
    ) -> list[float]:
        """A wrapper around `Plotext.sin`."""
        return plotext.sin(
            periods=periods,
            length=length,
            amplitude=amplitude,
            phase=phase,
            decay=decay,
        )

    @staticmethod
    def square(periods: int = 2, length: int = 200, amplitude: float = 1) -> list[int]:
        """A wrapper around `Plotext.square`."""
        return plotext.square(periods=periods, length=length, amplitude=amplitude)

    @staticmethod
    def colorize(
        string: str,
        fullground: Color | None = None,
        style: str | None = None,
        background: Color | None = None,
        show: bool = False,
    ) -> str:
        """A wrapper around `Plotet.colorize`."""
        return plotext.colorize(
            string,
            fullground=fullground,
            style=style,
            background=background,
            show=show,
        )

    @staticmethod
    def uncolorize(string: str) -> str:
        """A wrapper around `Plotext.uncolorize`."""
        return plotext.uncolorize(string)

    @staticmethod
    def transpose(data: list[list[Any]]) -> list[list[Any]]:
        """A wrapper around `Plotext.transpose`."""
        return plotext.transpose(data)

    platform = plotext.platform
    """The platform, as recognised by Plotext.

    One of either `unix` or `windows`.
    """

    def show(self) -> None:
        """Stub function. This should never be called within Textual."""

    def save_fig(
        self, path: str | None = None, append: bool = False, keep_colors: bool = False
    ) -> None:
        """Stub function. This should not be called within Textual."""
        # Actually... it might make sense to support this at some point and
        # might not be too tricky to handle. But for now this is about
        # allowing plotting within the application.
        del path, append, keep_colors


# Hoist the docstrings for the wrapper functions we've added above.
Plot.sin.__doc__ = plotext.sin.__doc__
Plot.square.__doc__ = plotext.square.__doc__
Plot.colorize.__doc__ = plotext.colorize.__doc__
Plot.uncolorize.__doc__ = plotext.uncolorize.__doc__
Plot.transpose.__doc__ = plotext.transpose.__doc__


def themes() -> tuple[str, ...]:
    """Get the list of available theme names.

    Returns:
        A tuple of the names of the themes defined in Plotext.
    """
    return get_args(ThemeName)


##############################################################################
# Here we patch in some textual-friendly themes. A Plotext theme is of the
# style:
#
# [canvas_color, axes_color, ticks_color, ticks_style, color_sequence]
#
# where color_sequence is a sequence of colours that will be used as plots
# are added (that is, if you plot 3 types of data and don't specify colours,
# this sequence will be used).


def _rgbify(color: Color) -> tuple[int, int, int] | str:
    """Force any Plotext colour into an RGB value.

    Args:
        color: The colour to convert.

    Returns:
        An RGB tuple for the colour.
    """
    if isinstance(color, str):
        # For some reason Plotext uses yellow and gold but never defines
        # them.
        #
        # TODO: Figure out a good version of those colours.
        return (
            color
            if color == "default"
            else _rgbify(
                get_color_code(
                    {"yellow": "orange", "gold": "orange+"}.get(color, color)
                )
            )
        )
    if isinstance(color, int):
        # A single integer; convert that into an RGB.
        return type1_to_type2_codes[color]
    # Must be an RGB already.
    return color


def _rgbify_theme(
    canvas_color: Color,
    axes_color: Color,
    ticks_color: Color,
    ticks_style: str,
    data_colors: list[Color],
) -> tuple[Color, Color, Color, str, list[Color]]:
    """Turn the given theme into a fully-RGB theme.

    Args:
        canvas_color: The canvas color.
        axes_color: The axes color.
        ticks_color: The ticks color.
        ticks_style: The styling for the ticks.
        data_colors: The sequence of colors for the various plots.

    Returns:
        The theme, made RGB-friendly.
    """
    return (
        _rgbify(canvas_color),
        _rgbify(axes_color),
        _rgbify(ticks_color),
        ticks_style,
        [_rgbify(data_color) for data_color in data_colors],
    )


_sequence = [
    (0, 130, 200),
    (60, 180, 75),
    (230, 25, 75),
    (255, 225, 25),
    (245, 130, 48),
    (145, 30, 180),
    (70, 240, 240),
    (240, 50, 230),
    (180, 225, 40),
    (250, 190, 212),
    (0, 128, 128),
    (220, 190, 255),
    (170, 110, 40),
    (235, 230, 180),
    (128, 0, 0),
    (170, 255, 195),
    (128, 128, 0),
    (255, 215, 180),
    (0, 0, 245),
    (128, 128, 128),
]
"""A sequence of colours for multiple plots.

Designed to work with either light or dark mode.
"""

# Make full-colour versions of the Plotext themes. In almost every case
# we'll follow the data sequence colours laid down by Plotext; but in a
# couple of cases we'll use our own curated set.
_themes["textual-default"] = list(
    _rgbify_theme("default", "default", "default", "default", _sequence)
)
_themes["textual-clear"] = list(
    _rgbify_theme(
        "default",
        "default",
        "default",
        "default",
        ["default"],
    )
)
_themes["textual-dark"] = _rgbify_theme(*_themes["dark"])
_themes["textual-dreamland"] = _rgbify_theme(*_themes["dreamland"])
_themes["textual-elegant"] = _rgbify_theme(*_themes["elegant"])
_themes["textual-girly"] = _rgbify_theme(*_themes["girly"])
_themes["textual-grandpa"] = _rgbify_theme(*_themes["grandpa"])
_themes["textual-matrix"] = _rgbify_theme(*_themes["matrix"])
_themes["textual-mature"] = _rgbify_theme(*_themes["mature"])
_themes["textual-pro"] = _themes["textual-default"]
_themes["textual-retro"] = _rgbify_theme(*_themes["retro"])
_themes["textual-sahara"] = _rgbify_theme(*_themes["sahara"])
_themes["textual-salad"] = _rgbify_theme(*_themes["salad"])
_themes["textual-scream"] = _rgbify_theme(*_themes["scream"])
_themes["textual-serious"] = _rgbify_theme(*_themes["serious"])
_themes["textual-windows"] = _rgbify_theme(*_themes["windows"])

# Now add a couple of extra themes that are Textual-specific.
_dark = DEFAULT_COLORS["dark"].generate()
_light = DEFAULT_COLORS["light"].generate()
_themes["textual-design-dark"] = [
    "default",
    "default",
    TextualColor.parse(_dark["accent-lighten-2"]).rgb,
    "bold",
    _sequence,
]
_themes["textual-design-light"] = [
    "default",
    "default",
    TextualColor.parse(_light["accent-darken-2"]).rgb,
    "bold",
    _sequence,
]



"""Main entry point for the library.

When run as:

```sh
$ python -m textual_plotext
```

it will show a demonstration of the library in action.
"""

# The following code borrows heavily from the Plotext readme files, and as
# such has some variables names and the like that would generally annoy
# pylint; so here we ask pylint to ease up for this particular file.
#
# Also, because it's common with Textual to introduce a property in
# `on_mount`, we ask pylint to ease up on that too.
#
# pylint:disable=invalid-name, attribute-defined-outside-init

from __future__ import annotations

import random
from datetime import datetime
from itertools import chain, cycle

from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Header, Label, Rule, TabbedContent, TabPane

from textual_plotext import PlotextPlot


class ExamplesPane(VerticalScroll):
    """Base class for a pane of examples."""

    DEFAULT_CSS = """
    ExamplesPane {
        scrollbar-gutter: stable;
    }

    ExamplesPane Label {
        margin: 2;
        width: 1fr;
        text-align: center;
    }
    """

    def examples(self, source: str, examples: list[PlotextPlot]) -> ComposeResult:
        """Provide the composed examples.

        Args:
            source: The source for the examples within the Plotext readme files.
            examples: A list of the example widgets.
        """
        yield Label(
            "Examples taken from https://github.com/piccolomo/"
            f"plotext/blob/master/readme/{source}.md"
        )
        for example in examples:
            yield example
            yield Rule()


class BasicPlots(ExamplesPane):
    """Examples from the basic section of the Plotext documentation."""

    class ScatterPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#scatter-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.scatter(self.plt.sin())
            self.plt.title("Scatter Plot")

    class LinePlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#line-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.plot(self.plt.sin())
            self.plt.title("Line Plot")

    class LogPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#log-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.plot(self.plt.sin(periods=2, length=10**4))
            self.plt.xscale("log")
            self.plt.yscale("linear")
            self.plt.grid(0, 1)
            self.plt.title("Logarithmic Plot")
            self.plt.xlabel("logarithmic scale")
            self.plt.ylabel("linear scale")
            # A slightly hacky workaround for what seems to be a bug in
            # Plotext itself when it comes to log scales. We force a build
            # of the data...
            _ = self.plt.build()
            # ...then put the scale back to linear so it doesn't try and
            # apply log again and again on each render.
            self.plt.xscale("linear")

    class StemPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#stem-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.plot(self.plt.sin(), fillx=True)
            self.plt.title("Stem Plot")

    class MultipleDataSets(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#multiple-data-sets"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.plot(self.plt.sin(), label="plot")
            self.plt.scatter(self.plt.sin(phase=-1), label="scatter")
            self.plt.title("Multiple Data Set")

    class MultipleAxesPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/basic.md#multiple-axes-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.plot(
                self.plt.sin(), xside="lower", yside="left", label="lower left"
            )
            self.plt.plot(
                self.plt.sin(2, phase=-1),
                xside="upper",
                yside="right",
                label="upper right",
            )
            self.plt.title("Multiple Axes Plot")

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        return self.examples(
            "basic",
            [
                self.ScatterPlot(),
                self.LinePlot(),
                self.LogPlot(),
                self.StemPlot(),
                self.MultipleDataSets(),
                self.MultipleAxesPlot(),
            ],
        )


class BarPlots(ExamplesPane):
    """Examples from the bar plots section of the Plotext documentation."""

    class VerticalBarPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/bar.md#vertical-bar-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            pizzas = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
            percentages = [14, 36, 11, 8, 7, 4]
            self.plt.bar(pizzas, percentages)
            self.plt.title("Most Favored Pizzas in the World")

    class HorizontalBarPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/bar.md#horizontal-bar-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            pizzas = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
            percentages = [14, 36, 11, 8, 7, 4]
            self.plt.bar(
                pizzas, percentages, orientation="horizontal", width=3 / 5
            )  # or in short orientation = 'h'
            self.plt.title("Most Favoured Pizzas in the World")

    class MultipleBarPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/bar.md#multiple-bar-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            pizzas = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
            male_percentages = [14, 36, 11, 8, 7, 4]
            female_percentages = [12, 20, 35, 15, 2, 1]
            self.plt.multiple_bar(
                pizzas, [male_percentages, female_percentages]
            )  # , labels = ["men", "women"])
            self.plt.title("Most Favored Pizzas in the World by Gender")

    class StackedBarPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/bar.md#stacked-bar-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            pizzas = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
            male_percentages = [14, 36, 11, 8, 7, 4]
            female_percentages = [12, 20, 35, 15, 2, 1]
            self.plt.stacked_bar(
                pizzas, [male_percentages, female_percentages]
            )  # , labels = ["men", "women"])
            self.plt.title("Most Favored Pizzas in the World by Gender")

    class HistogramPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/bar.md#histogram-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            l = 7 * 10**4
            data1 = [random.gauss(0, 1) for _ in range(10 * l)]
            data2 = [random.gauss(3, 1) for _ in range(6 * l)]
            data3 = [random.gauss(6, 1) for _ in range(4 * l)]
            bins = 60
            self.plt.hist(data1, bins, label="mean 0")
            self.plt.hist(data2, bins, label="mean 3")
            self.plt.hist(data3, bins, label="mean 6")
            self.plt.title("Histogram Plot")

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        return self.examples(
            "bar",
            [
                self.VerticalBarPlot(),
                self.HorizontalBarPlot(),
                self.MultipleBarPlot(),
                self.StackedBarPlot(),
                self.HistogramPlot(),
            ],
        )


class SpecialPlots(ExamplesPane):
    """Examples from the special plots section of the Plotext documentation."""

    class ErrorPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/special.md#error-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            l = 20
            ye = [random.random() for _ in range(l)]
            xe = [random.random() for _ in range(l)]
            data = self.plt.sin(length=l)
            self.plt.error(data, xerr=xe, yerr=ye)
            self.plt.title("Error Plot")

    class EventPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/special.md#event-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.date_form("H:M")
            times = self.plt.datetimes_to_string(
                [
                    datetime(
                        2022,
                        3,
                        27,
                        random.randint(0, 23),
                        random.randint(0, 59),
                        random.randint(0, 59),
                    )
                    for _ in range(100)
                ]
            )
            self.plt.event_plot(times)

    class StreamingDataPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/special.md#streaming-data"""

        def on_mount(self) -> None:
            """Set up the initial conditions for the 'streaming' data."""
            self.frame = 0
            self.plt.title("Streaming Data")
            self.set_interval(0.25, self.plot)

        def plot(self) -> None:
            """Plot the current frame of the stream."""
            self.plt.clear_data()
            self.plt.scatter(
                self.plt.sin(periods=2, length=1_000, phase=(2 * self.frame) / 50)
            )
            self.refresh()
            self.frame += 1

    class MatrixPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/special.md#matrix-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.title("Matrix Plot")

        def on_resize(self) -> None:
            """Remake the data when the size changes."""
            p = 1
            matrix = [
                [
                    (abs(r - self.size.height / 2) + abs(c - self.size.width / 2)) ** p
                    for c in range(self.size.width)
                ]
                for r in range(self.size.height)
            ]
            self.plt.clear_data()
            self.plt.matrix_plot(matrix)

    class ConfusionMatrix(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/special.md#confusion-matrix"""

        def on_mount(self) -> None:
            """Set up the plot."""
            l = 300
            actual = [random.randrange(0, 4) for _ in range(l)]
            predicted = [random.randrange(0, 4) for _ in range(l)]
            labels = ["Autumn", "Spring", "Summer", "Winter"]
            self.plt.cmatrix(actual, predicted, labels=labels)

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        return self.examples(
            "special",
            [
                self.ErrorPlot(),
                self.EventPlot(),
                self.StreamingDataPlot(),
                self.MatrixPlot(),
                self.ConfusionMatrix(),
            ],
        )


class DecoratorPlots(ExamplesPane):
    """Examples from the decorator plots section of the Plotext documentation."""

    class LinePlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/decorator.md#line-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.scatter(self.plt.sin())
            self.plt.title("Extra Lines")
            self.plt.vline(100, "magenta")
            self.plt.hline(0.5, "blue+")
            self.plt.plotsize(100, 30)

    class TextPlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/decorator.md#text-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            pizzas = ["Sausage", "Pepperoni", "Mushrooms", "Cheese", "Chicken", "Beef"]
            percentages = [14, 36, 11, 8, 7, 4]
            self.plt.bar(pizzas, percentages)
            self.plt.title("Labelled Bar Plot using Text()")
            for i, pizza in enumerate(pizzas):
                self.plt.text(
                    pizza,
                    x=i + 1,
                    y=percentages[i] + 1.5,
                    alignment="center",
                    color="red",
                )
            self.plt.ylim(0, 38)

    class ShapePlot(PlotextPlot):
        """https://github.com/piccolomo/plotext/blob/master/readme/decorator.md#shape-plot"""

        def on_mount(self) -> None:
            """Set up the plot."""
            self.plt.title("Shapes")
            self.plt.polygon()
            self.plt.rectangle()
            self.plt.polygon(sides=100)  # to simulate a circle
            self.plt.show()

    class PulsePlot(PlotextPlot):
        """I made this one up myself."""

        def on_mount(self) -> None:
            """Set up the pulsing polygon example."""
            self.plt.title("Pulse")
            self.steps = cycle(chain(range(3, 21), range(20, 2, -1)))
            self.set_interval(0.5, self.plot)

        def plot(self) -> None:
            """Plot a polygon with a changing number of sides."""
            self.plt.clear_data()
            self.plt.polygon(sides=next(self.steps))
            self.refresh()

    def compose(self) -> ComposeResult:
        return self.examples(
            "decorator",
            [self.LinePlot(), self.TextPlot(), self.ShapePlot(), self.PulsePlot()],
        )


class DemoApp(App[None]):
    """Demonstration application for the library."""

    CSS = """
    PlotextPlot {
        height: 40%
    }

    ShapePlot, PulsePlot {
        height: 100%;
    }
    """

    TITLE = "textual-plotext Demonstration"

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        yield Header()
        with TabbedContent():
            with TabPane("Basic Plots"):
                yield BasicPlots()
            with TabPane("Bar Plots"):
                yield BarPlots()
            with TabPane("Special Plots"):
                yield SpecialPlots()
            with TabPane("Decorator Plots"):
                yield DecoratorPlots()


if __name__ == "__main__":
    DemoApp().run()



"""Wrapper module for the Plotext _figure_class.

Designed to help type hint the methods in the class. See also the
accompanying pyi file for the types.
"""

from plotext._figure import _figure_class

__all__ = ["_figure_class"]
